﻿using System;
using SenseHat.Sensors;
using Windows.UI.Xaml.Controls;
using Windows.UI.Core;
using System.Threading.Tasks;

namespace SenseHat
{
    public sealed partial class MainPage : Page
    {
        private SensorReadings sensorReadings = new SensorReadings();

        private TemperatureAndPressureSensor temperatureAndPressureSensor = TemperatureAndPressureSensor.Instance;
        private HumidityAndTemperatureSensor humidityAndTemperatureSensor = HumidityAndTemperatureSensor.Instance;
        private InertialSensor inertialSensor = InertialSensor.Instance;
        private MagneticFieldSensor magneticFieldSensor = MagneticFieldSensor.Instance;

        public MainPage()
        {
            InitializeComponent();

            InitSensors();            
        }        

        private async void InitSensors()
        {
            // Temperature, and pressure sensor
            if (await temperatureAndPressureSensor.Initialize())
            {
                BeginTempAndPressureAcquisition();
            }

            // Humidity
            if (await humidityAndTemperatureSensor.Initialize())
            {
                BeginHumidityAcquisition();
            }

            // Inertial
            if (await inertialSensor.Initialize())
            {
                BeginAccelerationAndAngularSpeedAcquisition();
            }

            // Magnetic field
            if (await magneticFieldSensor.Initialize())
            {
                BeginMagneticFieldAcquisition();
            }
        }
        
        private void BeginTempAndPressureAcquisition()
        {
            const int msDelayTime = 40;

            BeginSensorReading(async () =>
            {
                // Get, and display sensors readings
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
                {
                    sensorReadings.Temperature = temperatureAndPressureSensor.GetTemperature();
                    sensorReadings.Pressure = temperatureAndPressureSensor.GetPressure();                    
                });
            }, msDelayTime);
        }

        private void BeginHumidityAcquisition()
        {
            const int msDelayTime = 80;

            BeginSensorReading(async () =>
            {
                // Get, and display sensors readings
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
                {
                    sensorReadings.Humidity = humidityAndTemperatureSensor.GetHumidity();                    
                });
            }, msDelayTime);
        }

        private void BeginAccelerationAndAngularSpeedAcquisition()
        {
            const int msDelayTime = 5;

            BeginSensorReading(async () =>
            {
                // Get, and display sensors readings
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
                {
                    sensorReadings.Accelerometer = inertialSensor.GetLinearAcceleration();
                    sensorReadings.Gyroscope = inertialSensor.GetAngularSpeed();                    
                });

            }, msDelayTime);
        }

        private void BeginMagneticFieldAcquisition()
        {
            const int msDelayTime = 13;

            BeginSensorReading(async () =>
            {
                // Get, and display sensors readings
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
                {
                    sensorReadings.Magnetometer = magneticFieldSensor.GetMagneticField();                    

                    var directionAngle = Helpers.MagneticFieldSensorHelper.GetDirectionAngle(sensorReadings.Magnetometer);
                    System.Diagnostics.Debug.WriteLine(directionAngle.ToString("F2"));
                });

            }, msDelayTime);
        }

        private void BeginSensorReading(Action periodicAction, int msDelayTime)
        {
            Task.Run(() =>
            {
                while (true)
                {
                    periodicAction();

                    Task.Delay(msDelayTime).Wait();
                }
            });
        }        
    }
}
