﻿using ArgumentValidation;
using MathNet.Numerics.IntegralTransforms;
using System;
using System.Numerics;

namespace FrequencyDistribution
{
    public static class SpectrumHelper
    {
        private const int binsCount = 8;
        private const double maxFrequency = 14000;

        public static double[] FourierMagnitude(short[] inputData)
        {
            Check.IsNull(inputData);

            var complexInput = ShortToComplexArray(inputData);

            Fourier.Forward(complexInput);

            return GetMagnitude(complexInput);
        }

        public static double[] Histogram(short[] inputData, double sampleRate)
        {
            Check.IsNull(inputData);

            // Windowing
            ApplyWindow(inputData);

            // FFT
            var fourierMagnitude = FourierMagnitude(inputData);

            // Db scale
            SetDbScale(fourierMagnitude);

            // Binning
            return GetHistogram(fourierMagnitude, sampleRate);
        }

        private static Complex[] ShortToComplexArray(short[] inputData)
        {
            int elementsCount = inputData.Length;
            var complexData = new Complex[elementsCount];

            for (int i = 0; i < elementsCount; i++)
            {
                complexData[i] = new Complex(inputData[i], 0.0d);
            }

            return complexData;
        }

        private static double[] GetMagnitude(Complex[] fft)
        {
            var magnitude = new double[fft.Length / 2];

            for (int i = 0; i < magnitude.Length; i++)
            {
                // Skip the DC component and frequencies 
                // higher than the Nyquist frequency
                magnitude[i] = fft[i + 1].Magnitude;
            }

            return magnitude;
        }

        private static void SetDbScale(double[] input)
        {
            for (int i = 0; i < input.Length; i++)
            {
                // Add epsilon to avoid infinity
                input[i] = 20.0 * Math.Log10(input[i] + double.Epsilon);
            }
        }

        private static void ApplyWindow(short[] inputData)
        {
            var window = MathNet.Numerics.Window.Hamming(inputData.Length);

            for (int i = 0; i < window.Length; i++)
            {
                inputData[i] = Convert.ToInt16(window[i] * inputData[i]);
            }
        }

        private static double[] GetHistogram(double[] dbFourierMagnitude, double sampleRate)
        {
            var histogram = new double[binsCount];

            var bins = GetFrequencyBins(sampleRate);

            var signalLength = 2 * dbFourierMagnitude.Length;
            var frequencyScale = Fourier.FrequencyScale(signalLength, sampleRate);

            for (int i = 0, frequencyIndex = 0; i < histogram.Length; i++)
            {
                var binWidth = 0;
                while (frequencyScale[frequencyIndex] <= bins[i])
                {
                    histogram[i] += dbFourierMagnitude[frequencyIndex];
                    binWidth++;

                    if (frequencyIndex++ == dbFourierMagnitude.Length - 1)
                    {
                        break;
                    }
                }

                histogram[i] = histogram[i] / binWidth;
            }

            return histogram;
        }

        private static double[] GetFrequencyBins(double sampleRate)
        {
            var bins = new double[binsCount];

            var startFrequency = sampleRate;

            for (int i = binsCount; i > 0; i--)
            {
                startFrequency /= 2;

                bins[i - 1] = Math.Min(startFrequency, maxFrequency);
            }

            return bins;
        }
    }
}
