﻿using System;
using OxyPlot;
using OxyPlot.Axes;
using Windows.UI.Xaml.Controls;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.UI.Core;
using SenseHatIO.SenseHatLedArray;
using SenseHatDisplay.Helpers;
using Windows.UI.Xaml.Navigation;

namespace FrequencyDistribution
{
    public sealed partial class MainPage : Page
    {
        private const int inputDataLength = 1024;

        private PlotModel sineWavePlotModel;
        private PlotModel fftPlotModel;

        private const string fileName = "audio.wav";

        private const double dbMinValue = 30;
        private const double dbMaxValue = 90;

        private MediaElement mediaElement;
        private bool isMediaElementPlaying;

        private double frequency;

        private LedArray ledArray;
        private bool isIoTPlatform = false;

        public double Frequency
        {
            get
            {
                return frequency;
            }
            set
            {
                frequency = value;
                ConfigureAxes();
                UpdatePlots();
            }
        }

        public MainPage()
        {
            InitializeComponent();

            InitializePlots();

            Frequency = 1;

            InitializeMediaElement();
        }

        protected override void OnNavigatedTo(NavigationEventArgs e)
        {
            base.OnNavigatedTo(e);

            InitializeLedArray();
        }

        private async void InitializeLedArray()
        {
            const byte address = 0x46;
            var device = await I2cHelper.GetI2cDevice(address);

            if (device != null)
            {
                ledArray = new LedArray(device);
                isIoTPlatform = true;
                ButtonProcessAudio_Click(null, null);
            }
        }

        private bool IsMediaElementPlaying
        {
            get
            {
                return isMediaElementPlaying;
            }
            set
            {
                isMediaElementPlaying = value;
                UpdateButtonCaption();
            }
        }

        private void InitializeMediaElement()
        {
            IsMediaElementPlaying = false;

            mediaElement = new MediaElement();
        }

        private void InitializePlots()
        {
            sineWavePlotModel = new PlotModel();
            fftPlotModel = new PlotModel();

            // Font size
            const int fontSize = 22;
            sineWavePlotModel.DefaultFontSize = fontSize;
            fftPlotModel.DefaultFontSize = fontSize;

            // Titles
            sineWavePlotModel.Title = "Time-domain";
            fftPlotModel.Title = "Frequency-domain";

            // Axes
            ConfigureAxes();
        }

        private void ConfigureAxes(bool histogram = false, int windowDataLength = inputDataLength)
        {
            sineWavePlotModel.Axes.Clear();
            fftPlotModel.Axes.Clear();

            sineWavePlotModel.Axes.Add(new LinearAxis()
            {
                Minimum = short.MinValue,
                Maximum = short.MaxValue,
                Position = AxisPosition.Left,
                Title = "Amplitude"
            });

            sineWavePlotModel.Axes.Add(new LinearAxis()
            {
                Minimum = 0,
                Maximum = windowDataLength,
                Position = AxisPosition.Bottom,
                Title = "Time"
            });

            fftPlotModel.Axes.Add(new LinearAxis()
            {
                Position = AxisPosition.Left,
                Title = "Amplitude",
                Minimum = histogram ? dbMinValue : double.NaN,
                Maximum = histogram ? dbMaxValue : double.NaN
            });

            fftPlotModel.Axes.Add(new LinearAxis()
            {
                Minimum = histogram ? -0.5 : 0,
                Maximum = histogram ? 7.5 : inputDataLength / 2,
                Position = AxisPosition.Bottom,
                Title = "Frequency"
            });
        }

        private void UpdatePlots()
        {
            var sineWave = PlotHelper.GenerateSineWave(inputDataLength, Frequency);
            var fourierMagnitude = SpectrumHelper.FourierMagnitude(sineWave);

            PlotHelper.AddLineSeries(sineWavePlotModel, sineWave, OxyColors.Blue);
            PlotHelper.AddLineSeries(fftPlotModel, fourierMagnitude, OxyColors.OrangeRed);
        }

        private async void ButtonProcessAudio_Click(object sender, Windows.UI.Xaml.RoutedEventArgs e)
        {
            if (!IsMediaElementPlaying)
            {
                // Get audio stream
                var audioStream = await StorageFileHelper.OpenLocalFile(fileName);

                // Get and parse buffer
                var waveBuffer = await StorageFileHelper.StreamToBuffer(audioStream);
                var waveData = new WaveData(waveBuffer.ToArray());

                // Adjust window length to sample rate
                var windowLength = AdjustWindowLength(waveData.SampleRate);

                // Update plots display ranges
                ConfigureAxes(true, windowLength);

                // Play the stream 
                IsMediaElementPlaying = true;

                mediaElement.SetSource(audioStream, audioStream.ContentType);
                mediaElement.Play();

                // Begin audio processing
                await Task.Run(() => {
                    DetermineAudioSpectrum(waveData, windowLength);
                });
            }
            else
            {
                mediaElement.Stop();

                IsMediaElementPlaying = false;
            }
        }

        private async void DetermineAudioSpectrum(WaveData waveData, int windowLength)
        {
            while (IsMediaElementPlaying)
            {
                var inputData = new short[windowLength];

                // Get MediaElement position
                int index = await GetWindowPosition(waveData.SampleRate, windowLength);

                if (index + windowLength < waveData.SamplesPerChannel)
                {
                    // Get current frame
                    Array.Copy(waveData.SamplesLeftChannel, index, inputData, 0, windowLength);

                    // Determine histogram
                    var hist = SpectrumHelper.Histogram(inputData, waveData.SampleRate);

                    if (isIoTPlatform)
                    {
                        ledArray.DrawHistogram(hist, dbMinValue, dbMaxValue);
                    }
                    else
                    {
                        // Plot input frame and histogram
                        PlotHelper.AddLineSeries(sineWavePlotModel, inputData, OxyColors.Blue);
                        PlotHelper.AddBarSeries(fftPlotModel, hist, OxyColors.Orange);
                    }
                }
                else
                {
                    // File ended
                    IsMediaElementPlaying = false;
                }
            }
        }

        private int AdjustWindowLength(double sampleRate)
        {
            var windowLength = 256;

            if (sampleRate >= 44100)
            {
                windowLength = 1024;
            }
            else if (sampleRate >= 22050)
            {
                windowLength = 512;
            }

            return windowLength;
        }

        private async Task<int> GetWindowPosition(double sampleRate, int windowLength)
        {
            int index = 0;

            await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
            {
                var position = mediaElement.Position.TotalSeconds * sampleRate;
                index = Convert.ToInt32(position) - windowLength / 2;
                index = Math.Max(index, 0);
            });

            return index;
        }

        private async void UpdateButtonCaption()
        {
            if (Dispatcher.HasThreadAccess)
            {
                string updatedContent = string.Empty;

                if (!isMediaElementPlaying)
                {
                    updatedContent = "Start audio processing";
                }
                else
                {
                    updatedContent = "Stop audio processing";
                }

                ButtonProcessAudio.Content = updatedContent;
            }
            else
            {
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, UpdateButtonCaption);
            }
        }
    }
}
